// Read/Write request processors for AIO167 driver

#include "stddcls.h"
#include "driver.h"

NTSTATUS FillInDeviceID(PDEVICE_EXTENSION pdx);
NTSTATUS ConfigureByID(PDEVICE_EXTENSION pdx);

PDEVICE_EXTENSION GlobalExtension[MAX_SLOTS_OPEN];

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchCreate(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchCreate
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

	// Claim the remove lock in Win2K so that removal waits until the
	// handle closes. Don't do this in Win98, however, because this
	// device might be removed by surprise with handles open, whereupon
	// we'll deadlock in HandleRemoveDevice waiting for a close that
	// can never happen because we can't run the user-mode code that
	// would do the close.

	NTSTATUS status;
	if ( win98 || ! pdx->bRealDevice )
		status = STATUS_SUCCESS;
	else
		status = IoAcquireRemoveLock(&pdx->RemoveLock, stack->FileObject);

	if (NT_SUCCESS(status))
	{						// okay to open
		if (InterlockedIncrement(&pdx->handles) == 1)
		{					// first open handle
		}					// first open handle
		KdPrint(("Opened handle, %d handle%s open\n", pdx->handles, pdx->handles-1 ? "s" : ""));
	}					// okay to open
	return CompleteRequest(Irp, status, 0);
	}							// DispatchCreate

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

void ForgetIRP(PIRP IRP)
{
	PAGED_CODE();
	//This function is evil. However, if we complete the IRP, it'll ret into a thread that no
	//longer exists. This way, it'll remain pending indefinitely, consuming resources
	//including nonpaged memory... but the system won't bugcheck.

	//It would be really nice of us to delete the IRP without completing it, but we don't
	//know how, and furthermore don't believe there's a way(since it wouldn't be necessary
	//without this "ret into a thread that no longer exists" bug).
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchClose(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchClose
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	if (InterlockedDecrement(&pdx->handles) == 0)
	{						// no more open handles
		if (pdx->bRealDevice && pdx->InterruptRequested && pdx->HasPendingIrp)
		{
			pdx->InterruptRequested = FALSE;
			pdx->HasPendingIrp = FALSE;
			ForgetIRP(pdx->PendingIrp);
			pdx->PendingIrp = NULL;
			pdx->InterruptOccurred = FALSE;
		}
	}						// no more open handles
	KdPrint(("Closed handle, %d handle%s open\n", pdx->handles, pdx->handles-1 ? "s" : ""));

	// Release the remove lock to match the acquisition done in DispatchCreate

	if ( pdx->bRealDevice && ! win98 )
		IoReleaseRemoveLock(&pdx->RemoveLock, stack->FileObject);

	return CompleteRequest(Irp, STATUS_SUCCESS, 0);
	}							// DispatchClose

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

VOID DoIOAction(PUCHAR Base, IOACTION Action)
{
	switch (Action.Operation)
	{
		case 'R':
			READ_PORT_UCHAR(Base + Action.Offset);
		break;
		case 'W':
			WRITE_PORT_UCHAR(Base + Action.Offset, Action.Data);
		break;
	}
}

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

VOID DpcForIsr(PKDPC Dpc, PDEVICE_OBJECT fdo, PIRP junk, PDEVICE_EXTENSION pdx)
{							// TrueDPC
	DoIOAction(pdx->portbase, pdx->Enable);

	if (pdx->InterruptRequested && pdx->HasPendingIrp)
	{
		pdx->InterruptRequested = FALSE;
		pdx->InterruptOccurred = FALSE;
		pdx->HasPendingIrp = FALSE;
		pdx->PendingIrp->IoStatus.Status = STATUS_NO_MORE_FILES;//I wish STATUS_SUCCESS worked...
		pdx->PendingIrp->IoStatus.Information = 2 * pdx->COSWords;
		IoSetCancelRoutine(pdx->PendingIrp, NULL);
		IoCompleteRequest(pdx->PendingIrp, IO_SOUND_INCREMENT);
	}
	else
	{
		pdx->InterruptOccurred = TRUE;
	}

}							// TrueDPC

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

BOOLEAN TrueISR(PDEVICE_EXTENSION pdx)
{							// TrueISR
	KdPrint((DRIVERNAME " - IRQ from %04LX?\n", pdx->DeviceID));
	UCHAR Data = 0xFF;
	if (pdx->IAm.Operation == 'R')
		Data = READ_PORT_UCHAR(pdx->portbase + pdx->IAm.Offset) & pdx->IAm.Data;
	//else if (pdx->IAm.Operation == 'Q')
	//	Data = READ_PORT_UCHAR(pdx->StatusBase + pdx->IAm.Offset) & pdx->IAm.Data;

	if (!Data) //then it wasn't me
		return FALSE;

	KdPrint((DRIVERNAME " - Yes!\n"));


	DoIOAction(pdx->portbase, pdx->Disable);

	if ( pdx->InterruptRequested && pdx->HasPendingIrp && pdx->COSWords )
		READ_PORT_BUFFER_USHORT(PUSHORT(pdx->portbase + 0x00), PUSHORT(pdx->PendingIrp->AssociatedIrp.SystemBuffer), pdx->COSWords)
	;

	//The DPC enables
	IoRequestDpc(pdx->DeviceObject, NULL, pdx);


	if (!pdx->busy)
		return TRUE;
	return TRUE;
}							// TrueISR

BOOLEAN OnInterrupt(PKINTERRUPT InterruptObject, PVOID corrupt)
{
	KdPrint((DRIVERNAME " - IRQ...\n"));
	BOOLEAN Anybody = FALSE;

	for (int I = MAX_SLOTS_OPEN - 1; I >= 0 && GlobalExtension[I]; --I)
		if (TrueISR(GlobalExtension[I]))
			Anybody = TRUE;

	return Anybody;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DumpRegDWord(PWCHAR ValueNom, ULONG Data)
{
	UNICODE_STRING KeyPath, ValueName;
	HANDLE Key;
	OBJECT_ATTRIBUTES Attr;
	NTSTATUS Status;

	RtlInitUnicodeString(&KeyPath, L"\\REGISTRY\\MACHINE\\Software\\ACCES\\AIO167");
	RtlInitUnicodeString(&ValueName, ValueNom);

	InitializeObjectAttributes(&Attr, &KeyPath, 0, NULL, NULL);
	Status = ZwOpenKey(&Key, KEY_SET_VALUE, &Attr);
	if (!NT_SUCCESS(Status))
		return Status;

	Status = ZwSetValueKey(Key,  &ValueName, 0, REG_DWORD, &Data, 4);

	ZwClose(Key);

	return Status;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS StartDevice(PDEVICE_OBJECT fdo, PCM_PARTIAL_RESOURCE_LIST raw, PCM_PARTIAL_RESOURCE_LIST translated)
	{							// StartDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	NTSTATUS status;

	// Identify the I/O resources we're supposed to use.

	ULONG vector;
	KIRQL irql;
	KINTERRUPT_MODE mode;
	KAFFINITY affinity;
	BOOLEAN irqshare;
	BOOLEAN gotinterrupt = FALSE;
	PHYSICAL_ADDRESS portbase;
	BOOLEAN gotport = FALSE;

	//ClaimStandardFailReason();

	if (!translated)
		return STATUS_DEVICE_CONFIGURATION_ERROR;		// no resources assigned??

	PCM_PARTIAL_RESOURCE_DESCRIPTOR resource = translated->PartialDescriptors;
	ULONG nres = translated->Count;
	for (ULONG i = 0; i < nres; ++i, ++resource)
	{						// for each resource
		switch (resource->Type)
		{					// switch on resource type

			case CmResourceTypePort:
				if ( ! gotport )
				{ //This is the main space.
					portbase = resource->u.Port.Start;
					pdx->nports = resource->u.Port.Length;
					pdx->mappedport = (resource->Flags & CM_RESOURCE_PORT_IO) == 0;
					gotport = TRUE;
				}
			break;

			case CmResourceTypeInterrupt:
				irql = (KIRQL) resource->u.Interrupt.Level;
				vector = resource->u.Interrupt.Vector;
				affinity = resource->u.Interrupt.Affinity;
				mode = (resource->Flags == CM_RESOURCE_INTERRUPT_LATCHED) ?
					Latched : LevelSensitive;
				irqshare = resource->ShareDisposition == CmResourceShareShared;
				gotinterrupt = TRUE;
			break;

			case CmResourceTypeMemory: //This is the memory-mapped PLX config space
			case CmResourceTypeDevicePrivate:
				break;
			default:
				KdPrint((DRIVERNAME " - Unexpected I/O resource type %d\n", resource->Type));
			break;
		}					// switch on resource type
	}						// for each resource

	// Verify that we got all the resources we were expecting

	if ( ! gotport )
	{
		KdPrint((DRIVERNAME " - Didn't get expected I/O resources\n"));
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}
	pdx->UnmappedBase = (USHORT)portbase.LowPart;

	if (pdx->mappedport)
	{						// map port address for RISC platform
		pdx->portbase = (PUCHAR) MmMapIoSpace(portbase, pdx->nports, MmNonCached);
		if ( ! pdx->portbase )
		{
			KdPrint((DRIVERNAME " - Unable to map port range %I64X, length %X\n", portbase, pdx->nports));
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}						// map port address for RISC platform
	else
	{
		pdx->portbase = (PUCHAR) portbase.QuadPart;
	}

	status = FillInDeviceID(pdx);
	if (!NT_SUCCESS(status))
	{
		KdPrint((DRIVERNAME " - Could not get virtual device ID!\n"));
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	status = ConfigureByID(pdx);
	if (!NT_SUCCESS(status))
	{
		KdPrint((DRIVERNAME " - Could not configure by ID!\n"));
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	pdx->GotIRQ = gotinterrupt;
	if (pdx->GotIRQ)
	{
		status = IoConnectInterrupt(&pdx->InterruptObject, OnInterrupt,
			(PVOID) pdx, NULL, vector, irql, irql, mode, irqshare, affinity, FALSE);
		if (!NT_SUCCESS(status))
		{
			KdPrint((DRIVERNAME " - IoConnectInterrupt failed - %X\n", status));
			if (pdx->mappedport)
			{
				MmUnmapIoSpace(pdx->portbase, pdx->nports);
			}
			pdx->portbase = NULL;
			return status;
		}
	}
	else
	{
		pdx->InterruptObject = NULL;
	}

	KdPrint((DRIVERNAME " - DeviceID = %04X, I/O base = %04X\n", pdx->DeviceID, pdx->portbase));

	return STATUS_SUCCESS;
	}							// StartDevice

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID StopDevice(IN PDEVICE_OBJECT fdo, BOOLEAN oktouch /* = FALSE */)
	{							// StopDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	if (pdx->InterruptObject)
		{						// disconnect interrupt

		DoIOAction(pdx->portbase, pdx->Disable);

		IoDisconnectInterrupt(pdx->InterruptObject);
		pdx->InterruptObject = NULL;
		}						// disconnect interrupt

	if (pdx->portbase && pdx->mappedport)
		MmUnmapIoSpace(pdx->portbase, pdx->nports);
	pdx->portbase = NULL;
	}							// StopDevice

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS FillInDeviceID(PDEVICE_EXTENSION pdx)
{
	WCHAR HID[256];
	ULONG Size;
	NTSTATUS Status = IoGetDeviceProperty(pdx->Pdo, DevicePropertyHardwareID, sizeof(HID), HID, &Size);
	if (!NT_SUCCESS(Status))
		return Status;

	enum DevSearchState {FindingD, FindingE, FindingV, Finding_, FindingDone}
		State = FindingD;

	ULONG I;
	for ( I = 0; I < Size && State != FindingDone; ++I )
		switch(HID[I])
		{
		case 'D': case 'd': if (State == FindingD) State = FindingE;    else State = FindingD; break;
		case 'E': case 'e': if (State == FindingE) State = FindingV;    else State = FindingD; break;
		case 'V': case 'v': if (State == FindingV) State = Finding_;    else State = FindingD; break;
		case '_':           if (State == Finding_) State = FindingDone; else State = FindingD; break;

		default:
			State = FindingD;
			break;
		}

	if ( State != FindingDone )
		return STATUS_OBJECT_NAME_INVALID;

	pdx->DeviceID = 0;
	for ( int J = 0; J >= 0; ++J, ++I )
	{
		switch(HID[I])
		{
		case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
			pdx->DeviceID = pdx->DeviceID * 0x10 + (HID[I] - '0');
			break;
		case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
			pdx->DeviceID = pdx->DeviceID * 0x10 + (HID[I] - 'A' + 10);
			break;
		case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
			pdx->DeviceID = pdx->DeviceID * 0x10 + (HID[I] - 'a' + 10);
			break;
		default:
			J = -16;
			//break;
		}
	}
	pdx->DeviceID |= 0x10000;
	return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS ConfigureByID(PDEVICE_EXTENSION pdx)
{
	USHORT FallbackBase;
	switch ( pdx->DeviceID )
	{
		case 0x1ECE8: //104-AIO16E
			FallbackBase = 0x300;
			pdx->IAm.Data = 0xF0;
			pdx->IAm.Offset = 0x13;
			pdx->IAm.Operation = 'R';
			pdx->Disable.Data = 0x00;
			pdx->Disable.Offset = 0x13;
			pdx->Disable.Operation = 'W';
			pdx->Clear.Data = 0;
			pdx->Clear.Offset = 0;
			pdx->Clear.Operation = 'N';
			pdx->Enable.Data = 0x02;
			pdx->Enable.Offset = 0x13;
			pdx->Enable.Operation = 'W';
		break;
		case 0x12230: //104-QUAD-8
			FallbackBase = 0x320;
			pdx->IAm.Data = 0xFF;
			pdx->IAm.Offset = 0x10;
			pdx->IAm.Operation = 'R';
			pdx->Disable.Data = 0;
			pdx->Disable.Offset = 0;
			pdx->Disable.Operation = 'N';
			pdx->Clear.Data = 0;
			pdx->Clear.Offset = 0;
			pdx->Clear.Operation = 'N';
			pdx->Enable.Data = 0;
			pdx->Enable.Offset = 0;
			pdx->Enable.Operation = 'N';
		break;
		default:
			FallbackBase = 0x160;
			pdx->IAm.Data = 0;
			pdx->IAm.Offset = 0;
			pdx->IAm.Operation = 'N';
			pdx->Disable.Data = 0;
			pdx->Disable.Offset = 0;
			pdx->Disable.Operation = 'N';
			pdx->Clear.Data = 0;
			pdx->Clear.Offset = 0;
			pdx->Clear.Operation = 'N';
			pdx->Enable.Data = 0;
			pdx->Enable.Offset = 0;
			pdx->Enable.Operation = 'N';
		//break;
	}
	/*
		if ( pdx->portbase == NULL ) 
		{
			pdx->UnmappedBase = FallbackBase;
			pdx->portbase = PUCHAR(pdx->UnmappedBase);
		}
	*/
	return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////

